import royalnet.commands as rc

try:
    import royalnet.serf.telegram as rst
except ImportError:
    rst = None

try:
    import royalnet.serf.discord as rsd
except ImportError:
    rsd = None

import royalnet.utils as ru
from ..tables.discord import Discord
from ..tables.telegram import Telegram


class RoyalnetsyncCommand(rc.Command):
    name: str = "royalnetsync"

    description: str = "Connect your chat account to Royalnet!"

    syntax: str = "{username} {password}"

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        async with data.session_acm() as session:
            author = await data.find_author(session=session, required=False)
            if author is not None:
                raise rc.UserError(f"This account is already connected to {author}!")

            username = args[0]
            password = " ".join(args[1:])

            user = await data.find_user(session=session, identifier=username)
            if user is None:
                raise rc.UserError("No such user.")
            try:
                successful = user.test_password(password)
            except ValueError:
                raise rc.UserError(f"User {user} has no password set!")
            if not successful:
                raise rc.InvalidInputError(f"Invalid password!")

            if rst is not None and isinstance(self.serf, rst.TelegramSerf):
                import telegram
                message: telegram.Message = data.message
                from_user: telegram.User = message.from_user
                TelegramT = self.alchemy.get(Telegram)
                tg_user: Telegram = await ru.asyncify(
                    session.query(TelegramT).filter_by(tg_id=from_user.id).one_or_none
                )
                if tg_user is None:
                    # Create
                    tg_user = TelegramT(
                        user=user,
                        tg_id=from_user.id,
                        first_name=from_user.first_name,
                        last_name=from_user.last_name,
                        username=from_user.username
                    )
                    session.add(tg_user)
                else:
                    # Edit
                    tg_user.first_name = from_user.first_name
                    tg_user.last_name = from_user.last_name
                    tg_user.username = from_user.username
                await ru.asyncify(session.commit)
                await data.reply(f"↔️ Account {tg_user} synced to {user}!")

            elif rsd is not None and isinstance(self.serf, rsd.DiscordSerf):
                import discord
                message: discord.Message = data.message
                ds_author: discord.User = message.author
                DiscordT = self.alchemy.get(Discord)
                ds_user: Discord = await ru.asyncify(
                    session.query(DiscordT).filter_by(discord_id=ds_author.id).one_or_none
                )
                if ds_user is None:
                    # Create
                    # noinspection PyProtectedMember
                    ds_user = DiscordT(
                        user=user,
                        discord_id=ds_author.id,
                        username=ds_author.name,
                        discriminator=ds_author.discriminator,
                        avatar_url=ds_author.avatar_url._url
                    )
                    session.add(ds_user)
                else:
                    # Edit
                    ds_user.username = ds_author.name
                    ds_user.discriminator = ds_author.discriminator
                    ds_user.avatar_url = ds_author.avatar_url
                await ru.asyncify(session.commit)
                await data.reply(f"↔️ Account {ds_user} synced to {ds_author}!")

            else:
                raise rc.UnsupportedError(f"Unknown interface: {self.serf.__class__.__qualname__}")



import re
from ...utils import escalating_odds
import datetime


# https://stackoverflow.com/questions/18621568/regex-replace-text-outside-html-tags
spooky_pattern = re.compile(r"o(?![^\[]*]|[^[]]*\[/)")
spooky_replacement = r'⊕'
spooky_replacement_url = r'[url=https://ryg.steffo.eu/#/2020/o]⊕[/url]'

url_pattern = re.compile(r"\[url=(.*?)](.*?)\[/url]")
url_replacement = r'\2 (\1)'


def escape(string: str) -> str:
    """Escape a string to be sent through Discord, and format it using RoyalCode.

    Warning:
        Currently escapes everything, even items in code blocks."""
    if escalating_odds(datetime.datetime(2020, 10, 31, 4, 0)):
        string = re.sub(spooky_pattern, spooky_replacement, string)

    string = string \
        .replace("*", "\\*") \
        .replace("_", "\\_") \
        .replace("`", "\\`") \
        .replace("[b]", "**") \
        .replace("[/b]", "**") \
        .replace("[i]", "_") \
        .replace("[/i]", "_") \
        .replace("[u]", "__") \
        .replace("[/u]", "__") \
        .replace("[c]", "`") \
        .replace("[/c]", "`") \
        .replace("[p]", "```") \
        .replace("[/p]", "```")

    string = re.sub(url_pattern, url_replacement, string)

    return string



import re
import typing

from .errors import InvalidInputError


class CommandArgs(list):
    """An interface to easily access the arguments of a command.

    Inherits from :class:`list`."""

    def __getitem__(self, item):
        """Access arguments as if they were a :class:`list`.

        Raises:
            InvalidInputError: if the requested argument does not exist.

        Examples:
            ::

                # /pasta spaghetti aldente
                >>> self[0]
                "spaghetti"
                >>> self[1]
                "aldente"
                >>> self[2]
                # InvalidInputError: Missing argument #3.
                >>> self[0:2]
                ["spaghetti", "aldente"]

        """
        if isinstance(item, int):
            try:
                return super().__getitem__(item)
            except IndexError:
                raise InvalidInputError(f'Missing argument #{item + 1}.')
        if isinstance(item, slice):
            try:
                return super().__getitem__(item)
            except IndexError:
                raise InvalidInputError(f'Cannot get arguments from #{item.start + 1} to #{item.stop + 1}.')
        raise ValueError(f"Invalid type passed to CommandArgs.__getattr__: {type(item)}")

    def joined(self, *, require_at_least=0) -> str:
        """Get the arguments as a space-joined string.

        Parameters:
            require_at_least: the minimum amount of arguments required.

        Raises:
            InvalidInputError: if there are less than ``require_at_least`` arguments.

        Returns:
            The space-joined string.

        Examples:
            ::

                # /pasta spaghetti aldente
                >>> self.joined()
                "spaghetti aldente"
                >>> self.joined(require_at_least=3)
                # InvalidInputError: Not enough arguments specified (minimum is 3).

            """
        if len(self) < require_at_least:
            raise InvalidInputError(f"Not enough arguments specified (minimum is {require_at_least}).")
        return " ".join(self)

    def match(self, pattern: typing.Union[str, typing.Pattern], *flags) -> typing.Sequence[typing.AnyStr]:
        """Match the :meth:`.joined` string to a :class:`re.Pattern`-like object.

        Parameters:
            pattern: The regex pattern to be passed to :func:`re.match`.

        Raises:
            InvalidInputError: if the pattern doesn't match.

        Returns:
            The matched groups, as returned by :func:`re.Match.groups`."""
        text = self.joined()
        match = re.match(pattern, text, *flags)
        if match is None:
            raise InvalidInputError("Invalid syntax.")
        return match.groups()

    def optional(self, index: int, default=None) -> typing.Optional[str]:
        """Get the argument at a specific index, but don't raise an error if nothing is found, instead returning the
        ``default`` value.

        Parameters:
            index: The index of the argument you want to retrieve.
            default: The value returned if the argument is missing.

        Returns:
            Either the argument or the ``default`` value, defaulting to ``None``.

        Examples:
            ::

                # /pasta spaghetti aldente
                >>> self.optional(0)
                "spaghetti"
                >>> self.optional(2)
                None
                >>> self.optional(2, default="carbonara")
                "carbonara"

        """
        try:
            return self[index]
        except InvalidInputError:
            return default



import datetime

import royalnet.constellation.api as rca
import royalnet.utils as ru
from ..tables.tokens import Token


class ApiAuthTokenStar(rca.ApiStar):
    path = "/api/auth/token/v1"

    parameters = {
        "get": {},
        "post": {
            "duration": "The duration in seconds of the new token."
        }
    }

    auth = {
        "get": True,
        "post": True,
    }

    tags = ["auth"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get information about the current login token."""
        token = await data.token()
        return token.json()

    @rca.magic
    async def post(self, data: rca.ApiData) -> ru.JSON:
        """Create a new login token for the authenticated user.

        Keep it secret, as it is basically a password!"""
        user = await data.user()
        try:
            duration = int(data["duration"])
        except ValueError:
            raise rca.InvalidParameterError("Duration is not a valid integer")
        new_token = Token.generate(self.alchemy, user, datetime.timedelta(seconds=duration))
        data.session.add(new_token)
        await data.session_commit()
        return new_token.json()



import asyncio as aio
import functools
import logging
import uuid
from typing import *

import websockets

from .broadcast import Broadcast
from .config import Config
from .errors import ConnectionClosedError, InvalidServerResponseError
from .package import Package
from .request import Request
from .response import Response, ResponseSuccess, ResponseFailure

log = logging.getLogger(__name__)


class PendingRequest:
    def __init__(self, *, loop: aio.AbstractEventLoop = None):
        if loop is None:
            self.loop = aio.get_event_loop()
        else:
            self.loop = loop
        self.event: aio.Event = aio.Event(loop=loop)
        self.data: Optional[dict] = None

    def __repr__(self):
        if self.event.is_set():
            return f"<{self.__class__.__qualname__}: {self.data.__class__.__name__}>"
        return f"<{self.__class__.__qualname__}>"

    def set(self, data):
        self.data = data
        self.event.set()


def requires_connection(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self.connect_event.wait()
        return await func(self, *args, **kwargs)

    return new_func


def requires_identification(func):
    @functools.wraps(func)
    async def new_func(self, *args, **kwargs):
        await self.identify_event.wait()
        return await func(self, *args, **kwargs)

    return new_func


class Link:
    def __init__(self, config: Config, request_handler, *,
                 loop: aio.AbstractEventLoop = None):
        self.config: Config = config
        self.nid: str = str(uuid.uuid4())
        self.websocket: Optional["websockets.WebSocketClientProtocol"] = None
        self.request_handler: Callable[[Union[Request, Broadcast]],
                                       Awaitable[Response]] = request_handler
        self._pending_requests: Dict[str, PendingRequest] = {}
        if loop is None:
            self._loop = aio.get_event_loop()
        else:
            self._loop = loop
        self.error_event: aio.Event = aio.Event(loop=self._loop)
        self.connect_event: aio.Event = aio.Event(loop=self._loop)
        self.identify_event: aio.Event = aio.Event(loop=self._loop)

    def __repr__(self):
        if self.identify_event.is_set():
            return f"<{self.__class__.__qualname__} (identified)>"
        elif self.connect_event.is_set():
            return f"<{self.__class__.__qualname__} (connected)>"
        elif self.error_event.is_set():
            return f"<{self.__class__.__qualname__} (error)>"
        else:
            return f"<{self.__class__.__qualname__} (disconnected)>"

    async def connect(self):
        """Connect to the :class:`Server` at :attr:`.config.url`."""
        log.debug(f"Connecting to Herald Server at {self.config.url}...")
        self.websocket = await websockets.connect(self.config.url, loop=self._loop)
        self.connect_event.set()
        log.debug(f"Connected!")

    @requires_connection
    async def receive(self) -> Package:
        """Recieve a :py:class:`Package` from the :py:class:`Server`.

        Raises:
            :exc:`ConnectionClosedError` if the connection is closed."""
        try:
            jbytes: bytes = await self.websocket.recv()
            package: Package = Package.from_json_bytes(jbytes)
        except websockets.ConnectionClosed:
            self.error_event.set()
            self.connect_event.clear()
            self.identify_event.clear()
            log.warning(f"Herald Server connection closed: {self.config.url}")
            # What to do now? Let's just reraise.
            raise ConnectionClosedError()
        if self.identify_event.is_set() and package.destination != self.nid:
            raise InvalidServerResponseError("Package is not addressed to this NetworkLink.")
        log.debug(f"Received package: {package}")
        return package

    @requires_connection
    async def identify(self) -> None:
        log.debug(f"Identifying...")
        await self.websocket.send(f"Identify {self.nid}:{self.config.name}:{self.config.secret}")
        response: Package = await self.receive()
        if not response.source == "<server>":
            raise InvalidServerResponseError("Received a non-service package before identification.")
        if "type" not in response.data:
            raise InvalidServerResponseError("Missing 'type' in response data")
        if response.data["type"] == "error":
            raise ConnectionClosedError(f"Identification error: {response.data['type']}")
        assert response.data["type"] == "success"
        self.identify_event.set()
        log.debug(f"Identified successfully!")

    @requires_identification
    async def send(self, package: Package):
        """Send a package to the :class:`Server`."""
        log.debug(f"Trying to send package: {package}")
        try:
            jbytes = package.to_json_bytes()
        except TypeError as e:
            log.fatal(f"Could not send package: {' '.join(e.args)}")
            raise
        await self.websocket.send(jbytes)
        log.debug(f"Sent package: {package}")

    @requires_identification
    async def broadcast(self, destination: str, broadcast: Broadcast) -> None:
        package = Package(broadcast.to_dict(), source=self.nid, destination=destination)
        await self.send(package)
        log.debug(f"Sent broadcast to {destination}: {broadcast}")

    @requires_identification
    async def request(self, destination: str, request: Request) -> Response:
        if destination.startswith("*"):
            raise ValueError("requests cannot have multiple destinations")
        package = Package(request.to_dict(), source=self.nid, destination=destination)
        request = PendingRequest(loop=self._loop)
        self._pending_requests[package.source_conv_id] = request
        await self.send(package)
        log.debug(f"Sent request to {destination}: {request}")
        await request.event.wait()
        if request.data["type"] == "ResponseSuccess":
            response: Response = ResponseSuccess.from_dict(request.data)
        elif request.data["type"] == "ResponseFailure":
            response: Response = ResponseFailure.from_dict(request.data)
        else:
            raise TypeError("Unknown response type")
        log.debug(f"Received from {destination}: {request} -> {response}")
        return response

    async def run(self):
        """Blockingly run the Link."""
        log.debug(f"Running link: {self.config.name}")
        if self.error_event.is_set():
            raise ConnectionClosedError("RoyalnetLinks can't be rerun after an error.")
        while True:
            if not self.connect_event.is_set():
                await self.connect()
            if not self.identify_event.is_set():
                await self.identify()
            package: Package = await self.receive()
            # Package is a response
            if package.destination_conv_id in self._pending_requests:
                request = self._pending_requests[package.destination_conv_id]
                request.set(package.data)
                continue
            # Package is a request
            elif package.data["msg_type"] == "Request":
                log.debug(f"Received request {package.source_conv_id}: {package}")
                response: Response = await self.request_handler(Request.from_dict(package.data))
                response_package: Package = package.reply(response.to_dict())
                await self.send(response_package)
                log.debug(f"Replied to request {response_package.source_conv_id}: {response_package}")
            # Package is a broadcast
            elif package.data["msg_type"] == "Broadcast":
                log.debug(f"Received broadcast {package.source_conv_id}: {package}")
                await self.request_handler(Broadcast.from_dict(package.data))



class ApiError(Exception):
    pass


class NotFoundError(ApiError):
    """404"""


class UnauthorizedError(ApiError):
    """401"""


class ForbiddenError(ApiError):
    """403"""


class BadRequestError(ApiError):
    """400"""


class ParameterError(BadRequestError):
    """400"""


class MissingParameterError(ParameterError):
    """400"""


class InvalidParameterError(ParameterError):
    """400"""


class MethodNotImplementedError(ApiError):
    """405"""


class UnsupportedError(MethodNotImplementedError):
    """405"""



from abc import ABCMeta

from .star import Star


class PageStar(Star, metaclass=ABCMeta):
    """A PageStar is a class representing a single route of the website (for example, ``/api/user/get``).

    To create a new website route you should create a new class inheriting from this class with a function overriding
    :meth:`.page`, :attr:`.path` and optionally :attr:`.methods`."""

    path: str = NotImplemented
    """The route of the star.

    Example:
        ::

            path: str = '/api/user/get'

    """

    @classmethod
    def methods(cls):
        """The HTTP methods supported by the Star, in form of a list.

        By default, a Star only supports the ``GET`` method, but more can be added.

        Example:
            ::

                methods: List[str] = ["GET", "POST", "PUT", "DELETE"]

        """
        return ["GET"]

    def __repr__(self):
        return f"<{self.__class__.__qualname__}: {self.path}>"



from royalnet.commands import *


class ExceptionEvent(HeraldEvent):
    name = "exception"

    def run(self, **kwargs):
        if not self.config["exc_debug"]:
            raise UserError(f"{self.__class__.__name__} is not enabled.")
        raise Exception(f"{self.name} event was called")



import royalnet.commands as rc
from ..tables import User


class RoyalnetaliasesCommand(rc.Command):
    name: str = "royalnetaliases"

    description: str = "Display your Royalnet aliases."

    syntax: str = ""

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        if name := args.optional(0) is not None:
            async with data.session_acm() as session:
                user = await User.find(alchemy=self.alchemy, session=session, identifier=name)
        else:
            user = await data.get_author(error_if_none=True)

        msg = [
            "👤 You currently have these aliases:",
            *list(map(lambda r: f"- {r}", user.aliases))
        ]

        await data.reply("\n".join(msg))



import random
import datetime
import logging

log = logging.getLogger(__name__)


def escalating_odds(target_day: datetime.datetime) -> bool:
    time_left: datetime.timedelta = target_day - datetime.datetime.now()
    log.debug(f"Time left is {time_left}")

    if time_left.total_seconds() <= 0:
        log.debug("Target has passed, autorolling a success")
        return True

    odds = 1 / (time_left.days + 1)
    result = random.uniform(0, 1) <= odds
    log.debug(f"Odds are {odds}, rolled a {result}")
    return result



import re
from typing import *
from ...utils import escalating_odds
import datetime


# https://stackoverflow.com/questions/18621568/regex-replace-text-outside-html-tags
spooky_pattern = re.compile(r"o(?![^\[]*]|[^[]]*\[/)")
spooky_replacement = r'⊕'
spooky_replacement_url = r'[url=https://ryg.steffo.eu/#/2020/o]⊕[/url]'

url_pattern = re.compile(r"\[url=(.*?)](.*?)\[/url]")
url_replacement = r'<a href="\1">\2</a>'


def escape(string: Optional[str]) -> Optional[str]:
    """Escape a string to be sent through Telegram (as HTML), and format it using RoyalCode.

    Warning:
        Currently escapes everything, even items in code blocks."""

    string = string.replace("<", "&lt;").replace(">", "&gt;")

    if escalating_odds(datetime.datetime(2020, 10, 31, 4, 0)):
        string = re.sub(spooky_pattern, spooky_replacement_url, string)

    string = string \
        .replace("[b]", "<b>") \
        .replace("[/b]", "</b>") \
        .replace("[i]", "<i>") \
        .replace("[/i]", "</i>") \
        .replace("[u]", "<b>") \
        .replace("[/u]", "</b>") \
        .replace("[c]", "<code>") \
        .replace("[/c]", "</code>") \
        .replace("[p]", "<pre>") \
        .replace("[/p]", "</pre>")

    string = re.sub(url_pattern, url_replacement, string)

    return string



import asyncio
import functools
import typing


async def asyncify(function: typing.Callable, *args, loop: typing.Optional[asyncio.AbstractEventLoop] = None, **kwargs):
    """Asyncronously run the function in a executor, allowing it to run asyncronously.

    Args:
        function: The function to call.
        args: The arguments to pass to the function.
        kwargs: The keyword arguments to pass to the function.
        loop: The loop to run the function in. If :const:`None`, run it in in the current event loop.

    Warning:
        The called function must be thread-safe!

    Warning:
        Calling a function this way might be significantly slower than calling its blocking counterpart!"""
    if not loop:
        loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, functools.partial(function, *args, **kwargs))



import royalnet.backpack.tables as rbt
import royalnet.constellation.api as rca
import royalnet.utils as ru


class ApiUserFindStar(rca.ApiStar):
    path = "/api/user/find/v1"

    parameters = {
        "get": {
            "alias": "One of the aliases of the user to get."
        }
    }

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get details about the Royalnet user with a certain alias."""
        user = await rbt.User.find(self.alchemy, data.session, data["alias"])
        if user is None:
            raise rca.NotFoundError("No such user.")
        return user.json()



from sqlalchemy import Column, \
    Integer, \
    String, \
    ForeignKey
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship


class Alias:
    __tablename__ = "aliases"

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"), primary_key=True)

    @declared_attr
    def alias(self):
        return Column(String, primary_key=True)

    @declared_attr
    def user(self):
        return relationship("User", backref="_aliases")

    def __init__(self, user, alias: str):
        self.user = user
        self.alias = alias.lower()

    def __repr__(self):
        return f"<Alias {str(self)}>"

    def __str__(self):
        return f"{self.alias}->{self.user_id}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['royalnet',
 'royalnet.alchemy',
 'royalnet.backpack',
 'royalnet.backpack.commands',
 'royalnet.backpack.events',
 'royalnet.backpack.stars',
 'royalnet.backpack.tables',
 'royalnet.backpack.utils',
 'royalnet.commands',
 'royalnet.constellation',
 'royalnet.constellation.api',
 'royalnet.herald',
 'royalnet.serf',
 'royalnet.serf.discord',
 'royalnet.serf.telegram',
 'royalnet.types',
 'royalnet.utils']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'dateparser>=0.7.2,<0.8.0', 'toml>=0.10.0,<0.11.0']

extras_require = \
{'alchemy_easy': ['sqlalchemy>=1.3.19,<2.0.0',
                  'psycopg2_binary>=2.8.6,<3.0.0',
                  'bcrypt>=3.2.0,<4.0.0'],
 'alchemy_hard': ['sqlalchemy>=1.3.19,<2.0.0',
                  'psycopg2>=2.8.6,<3.0.0',
                  'bcrypt>=3.2.0,<4.0.0'],
 'coloredlogs': ['coloredlogs>=10.0,<11.0'],
 'constellation': ['starlette>=0.12.13,<0.13.0',
                   'uvicorn>=0.10.7,<0.11.0',
                   'python-multipart>=0.0.5,<0.0.6'],
 'discord': ['discord.py>=1.3.1,<2.0.0', 'pynacl>=1.3.0,<2.0.0'],
 'herald': ['websockets>=8.1,<9.0'],
 'sentry': ['sentry_sdk>=0.13.2,<0.14.0'],
 'telegram': ['python_telegram_bot>=12.2.0,<13.0.0', 'urllib3>=1.25.10,<2.0.0']}

entry_points = \
{'console_scripts': ['royalnet = royalnet.__main__:run']}

setup_kwargs = {
    'name': 'royalnet',
    'version': '5.11.17',
    'description': 'A multipurpose bot and web framework',
    'long_description': '# `royalnet` [![PyPI](https://img.shields.io/pypi/v/royalnet.svg)](https://pypi.org/project/royalnet/)\n\nA framework for small Internet communities\n\n## About Royalnet\n\n`royalnet` is a Python framework composed of many interconnected services that may be useful to small Internet communities (gaming clans, university groups, etc).\n\n### [Serfs](royalnet/serf)\n\n_Serfs_ are services that allow Royalnet to respond to **chat commands** on multiple chat platforms.\n\nCommands using the Royalnet Serf API share their code between chat platforms: each serf will handle the specifics for their respective platform, preventing potential bugs due to code duplication!\n\n#### Supported chat platforms\n\n- [Telegram](https://core.telegram.org/bots)\n- [Discord](https://discordapp.com/developers/docs/)\n\nMore can easily be added by implementing a new serf!\n\n### [Alchemy](royalnet/alchemy)\n\nThe _Alchemy_ module allows all Royalnet services to use a **PostgreSQL database** with a [SQLAlchemy](https://www.sqlalchemy.org/) interface.\n\nThis allows the usage of a shared and easy-to-use ORM: Alchemy handles for you everything, from the creation of new tables to isolating in transactions the calls made from a Serf command to the database.\n\n### [Herald Network](royalnet/herald)\n\nAll Royalnet services can communicate with each other through the _Herald Network_, a websockets-based system allowing Remote Procedure Calls ("_events_") between services.\n \nFor example, in response to a Telegram message, the Telegram Serf can contact the Discord Serf to ask it to connect to voice chat in Discord. \n\nConnections between different hosts are possible too, even if they currently are unused by the Royalnet Launcher.\n\n### [Constellation](royalnet/constellation)\n\nThe Constellation service is a [Starlette](https://www.starlette.io )-based webserver that can supply dynamic pages ("_stars_") while being connected to the other parts of Royalnet through the Herald.\n\n#### APIs\n\nThe Constellation service also offers utilities for creating REST APIs as Python functions with `dict`s as inputs and outputs, leaving (de)serialization, transmission and eventually authentication to Royalnet.\n\n### Sentry\n\nRoyalnet can automatically report uncaught errors in all services to a [Sentry](https://sentry.io )-compatible server, while logging them in the console in development environments to facilitate debugging.\n\n### Packs\n\nNew Serf _commands_, Constellation _stars_, Herald _events_ and Alchemy _tables_ can be added to Royalnet through plugins called "Packs" that can be activated or deactivated on each single instance.\n\n#### Config\n\nEach pack can access only its individual section in the configuration file, preventing key conflicts (as long as the packs themselves don\'t share the same name). \n\n#### Template\n\nNew packs can be created starting from [this GitHub template](https://github.com/Steffo99/royalnet-pack-template).\n\n## Installing Royalnet\n\nTo install `royalnet`, run:\n```\npip install royalnet\n```\n\nTo install a specific module, run:\n```\npip install royalnet[MODULENAME]\n```\n\nTo install all `royalnet` modules, run:\n```\npip install royalnet[telegram,discord,matrix,alchemy_easy,bard,constellation,sentry,herald,coloredlogs]\n```\n\n## Documentation\n\nA work-in-progress documentation is available [here](https://gh.steffo.eu/royalnet/html).\n\n## Developing for Royalnet\n\nTo develop for `royalnet`, you need to have [Poetry](https://poetry.eustace.io/) installed on your PC.\n\nAfter you\'ve installed Poetry, clone the git repo with the command:\n\n```\ngit clone https://github.com/Steffo99/royalnet\n```\n\nThen enter the new directory:\n\n```\ncd royalnet\n```\n\nAnd finally install all dependencies and the package:\n\n```\npoetry install -E telegram -E discord -E matrix -E alchemy_easy -E constellation -E sentry -E herald -E coloredlogs\n```\n\n## Help!\n\nNeed help in anything Royalnet-related? [Open a issue on GitHub!](https://github.com/Steffo99/royalnet/issues/new)\n',
    'author': 'Stefano Pigozzi',
    'author_email': 'ste.pigozzi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Steffo99/royalnet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import *

import bcrypt
from sqlalchemy import Column, \
    Integer, \
    String, \
    LargeBinary, \
    inspect
from sqlalchemy.ext.declarative import declared_attr

import royalnet.utils as ru
from .aliases import Alias
from .roles import Role
from ...utils import JSON, asyncify


# noinspection PyAttributeOutsideInit
class User:
    __tablename__ = "users"

    @declared_attr
    def uid(self):
        return Column(Integer, unique=True, primary_key=True)

    @declared_attr
    def username(self):
        return Column(String, unique=True, nullable=False)

    @declared_attr
    def email(self):
        return Column(String, unique=True)

    @declared_attr
    def password(self):
        return Column(LargeBinary)

    @declared_attr
    def avatar_url(self):
        return Column(String)

    @classmethod
    async def find(cls, alchemy, session, identifier: Union[str, int]):
        if isinstance(identifier, str):
            alias = await ru.asyncify(session.query(alchemy.get(Alias)).filter_by(alias=identifier.lower()).one_or_none)
            if alias is None:
                return None
            else:
                return alias.user
        elif isinstance(identifier, int):
            return await ru.asyncify(session.query(alchemy.get(cls)).get, identifier)
        else:
            raise TypeError("alias is of an invalid type.")

    def json(self) -> JSON:
        return {
            "uid": self.uid,
            "username": self.username,
            "email": self.email,
            "password_set": self.password is not None,
            "avatar_url": self.avatar_url,
            "roles": self.roles,
            "aliases": self.aliases
        }

    def set_password(self, password: str) -> None:
        byte_password: bytes = bytes(password, encoding="UTF8")
        self.password = bcrypt.hashpw(byte_password, bcrypt.gensalt(14))

    def test_password(self, password: str) -> bool:
        if self.password is None:
            raise ValueError("No password is set")
        byte_password: bytes = bytes(password, encoding="UTF8")
        return bcrypt.checkpw(byte_password, self.password)

    @property
    def roles(self) -> list:
        # noinspection PyUnresolvedReferences
        return list(map(lambda a: a.role, self._roles))

    def add_role(self, alchemy, role: str) -> None:
        role = role.lower()
        session = inspect(self).session
        session.add(alchemy.get(Role)(user=self, role=role))

    async def delete_role(self, alchemy, role: str) -> None:
        role = role.lower()
        session = inspect(self).session
        role = await asyncify(session.query(alchemy.get(Role)).filter_by(user=self, role=role).one_or_none)
        if role is not None:
            session.delete(role)

    @property
    def aliases(self) -> list:
        # noinspection PyUnresolvedReferences
        return list(map(lambda a: a.alias, self._aliases))

    def add_alias(self, alchemy, alias: str) -> None:
        alias = alias.lower()
        session = inspect(self).session
        session.add(alchemy.get(Alias)(user=self, alias=alias))

    async def delete_alias(self, alchemy, alias: str) -> None:
        alias = alias.lower()
        session = inspect(self).session
        alias = await asyncify(session.query(alchemy.get(Alias)).filter_by(user=self, alias=alias).one_or_none)
        if alias is not None:
            session.delete(alias)

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.username}>"

    def __str__(self):
        return self.username
